/*  -*- Mode: java -*-*/

package tube.model.finder;

import java.awt.event.*;
import java.applet.*;
import java.awt.*;

import java.util.*;
import java.io.FileInputStream;
import java.io.File;
import java.io.StringBufferInputStream;
import java.io.InputStream;

EXTEND Macros;
EXTEND TopLevelMacros;

//SET GEN_HELP true; // doc_header macro will emit help text

DOC_HEADER 
/**
 * __tool_name__: Tube SPICE Modeling Curve Fitting Tool.
 *
 * Synopsis: __synopsis1__ 
 *           __synopsis_2_
 *
 * Version: __version__ __date__.
 *
 * Download link: http://www.dmitrynizh.com/__tool_jar__
 * Documentation:  __homeURL__
 *
 * This tool __tool_name__ is a part of a family of interactive
 * tools aiding in matching tube data and SPICE models by performing
 * interactive, iterative, visual parameter fitting while also
 * obtaining performance data such as Mu/Rp/Gm, loadline positioning,
 * aiding in finding optimal loading.
 * 
 * __tool_name__ supports the SPICE model equations
 * for Pentodes known as Improved Koren Models, with further
 * improvements - for screen current as suggested by Alexander Gurskii
 * in his Audioxpress article (feb-2011 issue) and for Child-Langmuir
 * grid current by dmitrynizh.
 *
 * COPYRIGHT, LICENSE.
 *
 * __copyright__
 *
 * __homeURL__
 *
 * Redistribution and use in source and/or binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * - Redistribution in source or binary form is provided for individual 
 *   non-commercial use only. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * LOG
 *
 * <from 2001 to 02-mar-2013: to be copied here from old log files>
 * mar- nov-13:      V2 of Toplevelmacros & PaintKIT log applies here.
 * 06-nov-13 V2.2.1  Fixed UL mode so that it derives screen V from Vp_qui, not EG2.
 * 10-Nov-13  2.3    various big & small layout and control dependency changes; color chooser fixes; enabled clipboard paste.
 *                   also, various fixes for cursor and mouse move/event glitches, re-layout continued.
 * 11-Nov-13  2.4    improved dependencies and layout
 * 15-nov-13  2.5,   reworked harmonics display using multicolor lines and % not dB
 * 17-nov-13  2.6    a Blend (linear interpolation by Vg)  of two SPICE models.
 * 08-nov-14  2.7    With fixed tabs, etc, enabled drag&drop & transforms on the background image.
 * 11-nov-14  2.7    Fixed DHP, level 2 grid models, a few typos, added back VCT param and control 
 * 11-nov-14  3.0    Switched to Alexander Gurskii screen current
 * 14-nov-14  3.1    Improved toop tips, added Ip+ig2 and Triode-strapped plots.
 * 15-nov-14  3.2    Added generation of triode model derived from pentode params.
 * 15-jan-15  -/-    Added dummy RE2 and E2
 * 16-jan-15  3.3    Added generation of simple linear slope (LAMBDA) taken from PaintIKIP and PaintMOSKIP variants
 * 16-jan-15  4.0    added toggle to switch between knees: atan (round, pent) and tahn (sharp, beam)
 * 28-jan-15  4.1    Experimental sigmoid (S-shaped) front slope and knee option (The error function formula)
 * 24-dec-16  4.2    Renamed,optimized the sigmoind parms (old names - see experiemntal PaintIKIP varaints)
 * 17-jan-17  4.3    New pop-up windows: options and help, better tooltips (Toplevelmacros features)
 *                   new BG image transforms, sliders that go negative (Toplevelmacros features), 
 *                   improved param dependencies
 * 29-jan-17  4.4    instead of atan/tanhn knee switch keep atan 'on' better top slope)
 *                   local, plate-to-grid NFB (Toplevelmacros feature, common for KIP and KIT)
 * 04-feb-17  4.5    small fixes and model optimizations, more help tooltips; compacted the controls more
 *                   improved param dependencies
 * 10-feb-17   4.6  more spice model improvements+optimizations
 * __date__ __version__  added power dissipation plot
 */  
final static String __date__ = "12-Aug-18"
  ,__version__    = "4.7"
  ,__synopsis1__  = "Generates Koren Improved Pentode SPICE models,"
  ,__synopsis_2_  = " with a number of extensions and enhancements (knee shape, tetrode kink)"
  ,__tool_name__   = "Paint KIP"
  ,__tool_jar__   = "paint_kip.jar"
  ,__copyright__  = "Copyright (c) 2001,2018 by Dmitry Nizhegorodov. All rights reserved."
  ,__homeURL__    = "http://www.dmitrynizh.com/tubeparams_image.htm"
  , K = "K", P = "P", G = "G", G2 = "G2";

EPILOGUE initDefaultWhiteBG {
  frameSize_Y += 108;
  frameSize_X -= 50;
  adjustXY(560, 590);
}

PROLOGUE initControls { 
  // vert Y pos offset of the top most slider
  modelParamsTopY = 172 - 3*20; 
  loadingParamsTopY = modelParamsTopY + modelParamsYSpan() + 3*20; // (* 6 21)
  
  if (blankBGMode) {
    showLoadLine = true;
  }
}

EPILOGUE initControls { 
//   report.setBounds(10, 6, 40, 24);
//   // report.setText("Make SPICE Model");
//   report.setBorder(null);
  int shiftx = 54;
  Rectangle r = togglePlotGrid.getBounds();
  togglePlotGrid.setBounds(r.x+shiftx, r.y, r.width, r.height);
  r = colorNext.getBounds();
  colorNext.setBounds(r.x+shiftx, r.y, r.width, r.height);
}

BEFORE_DRAW_MODEL {
  double Vg0 = 
    (isInputSnapped)
    ? (Vg_start - (((Vg_count-1.0)/2.0) * Vg_step))
    : doubleFromText(Vg_qui_TF.getText(), 1, Vg_qui, true);

  // update Vg_qui;
  Vg_qui = Vg0;

  Ip_qui = plateCurrent(Vp_qui, Vg_qui);
  if (showLoadLine) { 
    // abscissas = new double[7];
    // ordinates = new double[7];
  }
} 

// runs when model text is imported - [update params] or ctrl-v
PROLOGUE updateOrInitAllParams {
  // some shortened models as well as legacy ones do nto have these, pre-init
  KLAM(0);
  KLAMG(0);
  KNK(0);
  KNG(0);
  // KD KC etc?
}

WITH double knee_simple (double Ep) { 
  return isTanhKnee 
    ? (ATAN((Ep+KNEX)/KNEE)*Math.tanh(Ep/KNEE2))
    : ATAN(Ep/KNEE); 
}

// vgfrom: Inverse Of Koren Improved Triode formula.
PROLOGUE fromVpIpToVg {
  // first compensate Ip for ATAN(Ep/KVB)
  Ip -=  KLAM*Ep;  
  Ip /= (1 + KLAMG*Ep);
  // probem: we do not know Eg, and top_adjustments is complex.
  // solution: we use Vg_qui as a hack, and the result seems ok... the
  // error when mouse moves seems negligible for most 'normal' cases
  Ip /= top_adjustments(Ep, Vg_qui); 
  double atan = knee_simple(Ep);
  if (atan <  1e-3) atan = 1e-3;
  Ip = Ip/atan;

  double e_screen = isUL ? (Vp_qui + UL * (Ep - Vp_qui)) : EG2;
  Ep = e_screen;
}

WITH_HARMONIC_ANALYSIS isPP harmonicPlot

WITH_ANALYTICS showLoadLine loadLine isPP loadline_calcs loadline_calcs harmonicPlot K_asym

//---------------------------------------------------------------- ---------------      ---------------
//   PARAM type   label or null      variable ini  mult  x   y     boxlen  enabl  inc/dec  action
//                to hide the param                                        expr   % or x   statiement
//---------------------------------------------------------------- ---------------      --------------
WITH_PARAM double "Vg from"          Vg_start  0   1     0   40    4       true	  step(Vg_step)	    ;
WITH_PARAM double "decr"             Vg_step   5   1     82  40    4       true	  k(1.1)	    ;
WITH_PARAM int    "count"            Vg_count  9   1     140 40    4       true	  step(1)	    ;
WITH_PARAM double "Vp max"           Vp_MAX   750  1     210 40    5       true	  k(1.05)	    ;
WITH_PARAM double "Ip max"           Ip_MAX   250  0.001 300 40    5       true	  k(1.05)	    ;
WITH_PARAM double "CCG, pF"          CCG      3    1e-12 0   TR[5] 5       true	  k(1.2)	    ;
WITH_PARAM double "CGP, pF"          CGP      1.4  1e-12 106  TR[5] 5       true	  k(1.2)	    ;
WITH_PARAM double "CCP, pF"          CCP      1.9  1e-12 212 TR[5] 5       true	  k(1.2)	    ;
WITH_PARAM double "RFIL"             RFIL     10   1     320 TR[5] 5       isDHP  k(1.2)	    ;
// WITH_PARAM double "KVC"             KVC     2.57 1     0 TR[6] 5       true	  k(1.01)	    ;
								    
//----------------------------------------------------------------  -----
// alternative style
//WITH_PARAM double "Vg from/dec/#"    Vg_start 12 1     0     40 4 ;
//WITH_PARAM double ""                 Vg_step  12 1     128    40 4 ;
//WITH_PARAM int    ""                 Vg_count 11 1     164   40 4 ;
//WITH_PARAM double "Vp max"           Vp_MAX   600 1     207   40 5 ;
//WITH_PARAM double "Ip max"           Ip_MAX   200 0.001 290   40 5 ;
// WITH_PARAM int "resol"   Vp_count 300 1     200  40 5 ;


WITH void drawLoadlinePointProjections (Graphics g, double v, double i) {
  if (i < Ip_MAX) {
    g.drawLine(voltageToX(0), currentToY(i), voltageToX(v), currentToY(i));
    g.drawLine(voltageToX(v), currentToY(0), voltageToX(v), currentToY(i));

    if (blankBGMode) { // in blank BG mode, draw inside of the plot rectangle.
      g.drawString(onePlaceDF.format(v), voltageToX(v) + 4, Y_MAX-4);
      g.drawString(onePlaceDF.format(i*1000), X_MIN+4, currentToY(i)-2);
    } else { // when paint over image, draw outside of the plot rectangle.
      g.drawString(onePlaceDF.format(v), voltageToX(v) - 20, Y_MAX+14);
      g.drawString(onePlaceDF.format(i*1000), X_MIN-40, currentToY(i)+4);
    }
  }
} 

WITH String gridModelParams () {
  return gridLevel2 ? advancedGridModelParams() : "RGI=" + RGI;
}

WITH String gridModelCir () {
  return gridLevel2 ? advancedGridModelCir() : 
    "R1   G 5  {RGI}  ; FOR GRID CURRENT   \n" +
    "D3   5 K  DX     ; FOR GRID CURRENT } \n" +
    ".MODEL DX D(IS=1N RS=1 CJO=10PF TT=1N)";
}

WITH String plateAndScreenCurrentCir (String K, String SUFF) {
  return advSigmoid 
    ? (addKink 
       ? advancedPlateAndScreenCirWithKink()
       : advancedPlateAndScreenCirNoKink())
    : (addKink 
       ? korenPlateAndGurskiiScreenAndOptSlopesWithKinksCir() 
       : korenPlateAndGurskiiScreenAndOptSlopesCir(K, SUFF));
}

WITH_SPICE_SUBCKT
/** 
  * **** ?NAME? ******************************************
  * * Created on ?DATE? using paint_kip.jar
  * * www.dmitrynizh.com/tubeparams_image.htm
  * * Plate Curves image file: ?plateCurvesImageFilePath?
  * * Data source link: <plate curves URL>
  * *----------------------------------------------------------------------------------
  * .SUBCKT ?NAME? P G2 G K ; LTSpice tetrode.asy pinout
  * * .SUBCKT ?NAME? P G K G2 ; Koren Pentode Pspice pinout
  * + PARAMS: MU=?MU? KG1=?KG1? KP=?KP? KVB=?KVB? VCT=?VCT? EX=?EX? KG2=?KG2? KNEE=?KNEE? KVC=?KVC?
  * + ?(slopeParamsByNeed())? ?(maybeSecondKneeParams())? ?(screenModelParams())? KNK=?KNK? KNG=?KNG? ?(otherKinkParams())?
  * + CCG=?CCG?P CGP=?CGP?P CCP=?CCP?P ?(gridModelParams())?
  * * Vp_MAX=?Vp_MAX? Ip_MAX=?Ip_MAX? Vg_step=?Vg_step? Vg_start=?Vg_start? Vg_count=?Vg_count?
  * * X_MIN=?X_MIN? Y_MIN=?Y_MIN? X_SIZE=?X_SIZE? Y_SIZE=?Y_SIZE? FSZ_X=?FSZ_X? FSZ_Y=?FSZ_Y? XYGrid=?XYGrid?
  * * Rp=?Rp? Vg_ac=?Vg_ac? P_max=?P_max? Vg_qui=?Vg_qui? Vp_qui=?Vp_qui? 
  * * showLoadLine=?showLoadLine? showIp=?showIp? isDHP=?isDHP? isPP=?isPP? isAsymPP=?isAsymPP? isUL=?isUL? showDissipLimit=?showDissipLimit? 
  * * showIg1=?showIg1? isInputSnapped=?isInputSnapped? addLocalNFB=?addLocalNFB?
  * * XYProjections=?XYProjections? harmonicPlot=?harmonicPlot? dissipPlot=?dissipPlot? 
  * * UL=?UL? EG2=?EG2? gridLevel2=?gridLevel2? addKink=?addKink? isTanhKnee=?isTanhKnee? advSigmoid=?advSigmoid? 
  * *----------------------------------------------------------------------------------
  * ?(korenTriodeVoltSrcs(K,""))?
  * ?(plateAndScreenCurrentCir(K, ""))?
  * RCP  P K  1G     ; FOR CONVERGENCE
  * C1   K G  {CCG}  ; CATHODE-GRID 1
  * C2   G P  {CGP}  ; GRID 1-PLATE
  * C3   K P  {CCP}  ; CATHODE-PLATE
  * ?(gridModelCir())?
  * .ENDS
  * *$
  * 
  */
SPICE genKorenImprovedSpiceModel () { 
  doLIN(MU,KG1,KP,KVB,VCT,EX,KNEE);
}

WITH_SPICE_SUBCKT
/** 
  * * The following triode model is derived from pentode model, see above.
  * * In the triode model, all spice parameters come directly from the pentode model, except for Kg1,
  * * which for triode-strapped pentodes is derived from pentode's Kg1, Kg2 and Kvc as
  * *
  * *    4Kg1Kg2 / ((2Kvc-Pi)(2Kg1+PiKg2))
  * 
  * **** ?NAME? ******************************************
  * * Created on ?DATE? using ?TOOL? 
  * * www.dmitrynizh.com/tubeparams_image.htm
  * * Plate Curves image file: ?plateCurvesImageFilePath?
  * * Data source link: <plate curves URL>
  * *----------------------------------------------------------------------------------
  * .SUBCKT TRIODE_?NAME? 1 2 3 ; Plate Grid Cathode
  * + PARAMS: CCG=?CCG?P  CGP=?CGP?P CCP=?CCP?P RGI=?RGI?
  * + MU=?MU? KG1=?(4*KG1*KG2/((2*KVC-3.1416)*(2*KG1+3.1416*KG2)))? KP=?KP? KVB=?KVB? VCT=?VCT? EX=?EX? 
  * * Vp_MAX=?Vp_MAX? Ip_MAX=?Ip_MAX? Vg_step=?Vg_step? Vg_start=?Vg_start? Vg_count=?Vg_count?
  * * Rp=?Rp? Vg_ac=?Vg_ac? P_max=?P_max? Vg_qui=?Vg_qui? Vp_qui=?Vp_qui?
  * * X_MIN=?X_MIN? Y_MIN=?Y_MIN? X_SIZE=?X_SIZE? Y_SIZE=?Y_SIZE? FSZ_X=?FSZ_X? FSZ_Y=?FSZ_Y? XYGrid=?XYGrid?
  * * showLoadLine=?showLoadLine? showIp=?showIp? isDHT=?isDHP? isPP=?isPP? isAsymPP=?isAsymPP? showDissipLimit=?showDissipLimit? 
  * * showIg1=?showIg1? gridLevel2=?gridLevel2? isInputSnapped=?isInputSnapped?  
  * * XYProjections=?XYProjections? harmonicPlot=?harmonicPlot? dissipPlot=?dissipPlot? 
  * *----------------------------------------------------------------------------------
  * E1 7 0 VALUE={V(1,3)/KP*LOG(1+EXP(KP*(1/MU+(VCT+V(2,3))/SQRT(KVB+V(1,3)*V(1,3)))))} 
  * RE1 7 0 1G  ; TO AVOID FLOATING NODES
  * G1 1 3 VALUE={(PWR(V(7),EX)+PWRS(V(7),EX))/KG1} 
  * RCP 1 3 1G   ; TO AVOID FLOATING NODES
  * C1 2 3 {CCG} ; CATHODE-GRID 
  * C2 2 1 {CGP} ; GRID=PLATE 
  * C3 1 3 {CCP} ; CATHODE-PLATE 
  * D3 5 3 DX ; POSITIVE GRID CURRENT 
  * R1 2 5 {RGI} ; POSITIVE GRID CURRENT 
  * .MODEL DX D(IS=1N RS=1 CJO=10PF TT=1N) 
  * .ENDS 
  * *$
  * 
  */
SPICE genKorenImprovedTriodeSpiceModel () { 
  doLIN(MU,KG1,KP,KVB,VCT,EX,KNEE);

}


WITH_SPICE_SUBCKT
/** 
  * **** ?NAME? Composite DHP *****************************************
  * * Created on ?DATE? using paint_kip.jar
  * * www.dmitrynizh.com/tubeparams_image.htm
  * * Plate Curves image file: ?plateCurvesImageFilePath?
  * * Data source link: <plate curves URL>
  * *----------------------------------------------------------------------------------
  * .SUBCKT DHP_?NAME? P G K1 K2 G2 ; Direct Filament Koren Pinout
  * + PARAMS: CCG=?CCG?P  CGP=?CGP?P CCP=?CCP?P RGI=?RGI? RFIL=?RFIL?
  * + MU=?MU? KG1=?KG1? KP=?KP? KVB=?KVB? KVC=?KVC? VCT=?VCT? EX=?EX? KG2=?KG2? KNEE=?KNEE?  
  * * Vp_MAX=?Vp_MAX? Ip_MAX=?Ip_MAX? Vg_step=?Vg_step? Vg_start=?Vg_start? Vg_count=?Vg_count?
  * * X_MIN=?X_MIN? Y_MIN=?Y_MIN? X_SIZE=?X_SIZE? Y_SIZE=?Y_SIZE? FSZ_X=?FSZ_X? FSZ_Y=?FSZ_Y? XYGrid=?XYGrid?
  * * Rp=?Rp? Vg_ac=?Vg_ac? P_max=?P_max? Vg_qui=?Vg_qui? Vp_qui=?Vp_qui? UL=?UL? EG2=?EG2? gridLevel2=?gridLevel2? 
  * * showLoadLine=?showLoadLine? showIp=?showIp? isDHP=?isDHP? isPP=?isPP? isAsymPP=?isAsymPP? isUL=?isUL? showDissipLimit=?showDissipLimit? showIg1=?showIg1? isInputSnapped=?isInputSnapped? XYProjections=?XYProjections? harmonicPlot=?harmonicPlot? dissipPlot=?dissipPlot? 
  * *----------------------------------------------------------------------------------
  * RFIL_LEFT    K1  33 {RFIL/4} ;; RFIL = K1-RFIL/4-33-RFIL/4-34-RFIL/4-35-RFIL/4-K2
  * RFIL_MIDDLE1 33  34  {RFIL/4}
  * RFIL_MIDDLE2 34  35 {RFIL/4}
  * RFIL_RIGHT   35  K2 {RFIL/4}
  * ?(korenTriodeVoltSrcs("33","1"))?
  * ?(korenTriodeVoltSrcs("35","2"))?
  * ?(plateAndScreenCurrentCir("33", "1"))?
  * ?(plateAndScreenCurrentCir("35", "2"))?
  * C1 34 G {CCG} ; CATHODE-GRID 
  * C2 G  P {CGP} ; GRID=PLATE 
  * C3 34 P {CCP} ; CATHODE-PLATE 
  * D3 6 K1 DX ; FOR GRID CURRENT 
  * D4 7 K2 DX ; FOR GRID CURRENT 
  * RG1 G 6 {2*RGI} ; FOR GRID CURRENT 
  * RG2 G 7 {2*RGI} ; FOR GRID CURRENT 
  * .MODEL DX D(IS=1N RS=1 CJO=10PF TT=1N) 
  * .ENDS 
  * *$
  * 
  */
SPICE genDHPKorenImprovedSpiceModel () { 
  doLIN(MU,KG1,KP,KVB,VCT,EX,KNEE);
}

WITH_SPICE_SUBCKT
/**RE1?(SUFF)?  7?(SUFF)? 0  1G    ; DUMMY SO NODE 7?(SUFF)? HAS 2 CONNECTIONS
  * E1?(SUFF)?   7?(SUFF)? 0  VALUE=  ; E1?(SUFF)? BREAKS UP LONG EQUATION FOR G1?(SUFF)?.
  * +{V(G2,?(K)?)/KP*LOG(1+EXP((1/MU+(VCT+V(G,?(K)?))/SQRT(KVB+V(G2,?(K)?)*V(G2,?(K)?)))*KP))}
  * RE2?(SUFF)?  6?(SUFF)? 0  1G    ; DUMMY SO NODE 6?(SUFF)? HAS 2 CONNECTIONS
  * E2?(SUFF)?  6?(SUFF)? 0  VALUE={(PWR(V(7?(SUFF)?),EX)+PWRS(V(7?(SUFF)?),EX))} ; Kg1 times KIT current */
SPICE korenTriodeVoltSrcs (String K, String SUFF) {}

WITH_SPICE_SUBCKT /** KLAM=?KLAM?*/  SPICE KLAM_param () { }
WITH_SPICE_SUBCKT /** KLAMG=?KLAMG?*/SPICE KLAMG_param() { }

WITH String slopeParamsByNeed () {
  return ((KLAM == 0 ? "" : KLAM_param()) + (KLAMG == 0 ? "" : KLAMG_param())).trim();
}

WITH String maybeSecondKneeParams () {
  return isTanhKnee ? secondKneeParams() : "";
}

WITH_SPICE_SUBCKT /**KNEE2=?KNEE2? KNEX=?KNEX?*/SPICE secondKneeParams () {}

WITH_SPICE_SUBCKT
/**VGOFF=?VGOFF? IGA=?IGA? IGB=?IGB? IGC=?IGC? IGEX=?IGEX?*/
SPICE advancedGridModelParams () { }

WITH_SPICE_SUBCKT
/**RE23 G 0 1G 
 * GG G K VALUE={(IGA+IGB/(IGC+V(P,K)))*(MU/KG1)*
 * +(PWR(V(G,K)-VGOFF,IGEX)+PWRS(V(G,K)-VGOFF,IGEX))} */ 
SPICE advancedGridModelCir () { }

WITH String screenModelParams () { 
  return 
    (advSigmoid 
     ? (advancedScreenModelParams() + 
        ((KVBGI == 0.001 && KB1 == 2 && KB2 == 2 && KB3 == 2 & KB4 == 0.5)
         ? "" : (" " + advancedScreenModelParamsPart2()))) 
     : "").trim();
}

WITH_SPICE_SUBCKT
/**KD=?KD? KC=?KC? KR1=?KR1? KR2=?KR2? KVBG=?KVBG?*/
SPICE advancedScreenModelParams () { }

WITH_SPICE_SUBCKT
/**KB1=?KB1? KB2=?KB2? KB3=?KB3? KB4=?KB4? KVBGI=?KVBGI?*/
SPICE advancedScreenModelParamsPart2 () { }


WITH String otherKinkParams () { return addKink == false ? "" : advancedKinkParams(); }
WITH_SPICE_SUBCKT 
/**KNPL=?KNPL? KNSL=?KNSL? KNPR=?KNPR? KNSR=?KNSR?*/
SPICE advancedKinkParams () { }


WITH_SPICE_SUBCKT /**ATAN(V(P,?(K)?)/KNEE)*/          SPICE atanKnee (String P, String K) { }
WITH_SPICE_SUBCKT /**ATAN((V(P,?(K)?)+KNEX)/KNEE)*TANH(V(P,?(K)?)/KNEE2)*/SPICE atanAndTanhKnee (String P, String K) { }
WITH String trig_knee_coeff_expr (String K) { return isTanhKnee ? atanAndTanhKnee(P, K) : atanKnee(P, K); }

WITH_SPICE_SUBCKT
/**G1?(SUFF)?   P ?(K)?  VALUE={V(6?(SUFF)?)/KG1*?(trig_knee_coeff_expr(K))??(optionalPlateSlope(K))?}
 * * Alexander Gurskii screen current, see audioXpress 2/2011
 * RE4?(K)? 4?(K)? ?(K)? 1G ; Dummy, per Alex request
 * E4?(K)? 4?(K)?  4 VALUE={0} ; Dummy, per Alex request
 * G4?(K)? 4?(K)? ?(K)? VALUE={V(6?(SUFF)?)/KG2*(KVC-?(trig_knee_coeff_expr(K))?)?(optionalScreenSlope(K))?}*/
SPICE korenPlateAndGurskiiScreenAndOptSlopesCir (String K, String SUFF) { 
}

// this has output range of 1-0.3pi=0.06 to 1
WITH String kink_coeff_expr = 
  "LIMIT(KNK-V(G,K)*KNG,0,0.3)*(-ATAN((V(P,K)-KNPL)/KNSL)+ATAN((V(P,K)-KNPR)/KNSR))";


WITH_SPICE_SUBCKT
/**RE21 21 0 1
 * E21  21 0 VALUE={V(6)/KG1*?(trig_knee_coeff_expr(K))?} ; Ip with knee but no slope and no kink
 * RE22 22 0 1 ; E22: kink curr deviation for plate
 * E22  22 0 VALUE={V(21)*?(kink_coeff_expr)?} 
 * G1   P  K  VALUE={V(21)?(optionalPlateSlope(K))? + V(22)}
 * * Alexander Gurskii screen current, see audioXpress 2/2011, with slope and kink added
 * RE43 43 K 1G ; Dummy
 * E43  43 G2 VALUE={0} ; Dummy
 * G2   43 K  VALUE={V(6)/KG2*(KVC-?(trig_knee_coeff_expr(K))?)?(optionalScreenSlope(K))?-V(22)}*/
SPICE korenPlateAndGurskiiScreenAndOptSlopesWithKinksCir () { 
}

WITH String advancedPlateAndScreenCirV9 () {
  return (KVBGI == 0.001 && KB1 == 2 && KB2 == 2 && KB3 == 2 & KB4 == 0.5)
    ? advancedPlateAndScreenCirV9Case1()
    : advancedPlateAndScreenCirV9Case2();
}

WITH_SPICE_SUBCKT
/**?(advancedPlateAndScreenCirV9())?
 * G1   P K  VALUE={V(6)/KG1*V(9)?(optionalPlateSlope(K))?}
 * G2   G2 K  VALUE={V(6)/KG2*(KVC-V(9))?(optionalScreenSlope(K))?} */
SPICE advancedPlateAndScreenCirNoKink () { 
}

WITH_SPICE_SUBCKT
/**?(advancedPlateAndScreenCirV9())?
 * RE21 21 0 1
 * E21  21 0 VALUE={V(6)/KG1*V(9)} ; Ip with knee but no slope and no kink
 * RE22 22 0 1 ; E22: kink curr deviation for plate
 * E22  22 0 VALUE={V(21)*?(kink_coeff_expr)?} 
 * G1   P K  VALUE={V(21)?(optionalPlateSlope(K))? + V(22)}
 * G2   G2 K  VALUE={V(6)/KG2*(KVC-V(9))?(optionalScreenSlope(K))? - V(22)} */
SPICE advancedPlateAndScreenCirWithKink () { 
}

WITH_SPICE_SUBCKT
/**E4   8 0  VALUE={PWR(V(1,3)/KNEE/(V(6)*KVBG),2)}
 * E5   9 0  VALUE={SQRT(1-EXP(-V(8)*(KC+KR1*V(8))/(KD+KR2*V(8))))*1.5708}
 * RE4  8 0  1
 * RE5  9 0  1 */
SPICE advancedPlateAndScreenCirV9Case1 () { 
}

WITH_SPICE_SUBCKT
/**E4   8 0  VALUE={V(P,K)/KNEE/(KVBGI+V(6)*KVBG)}
 * E5  81 0  VALUE={PWR(V(8),KB1)}
 * E6  82 0  VALUE={PWR(V(8),KB2)}
 * E7  83 0  VALUE={PWR(V(8),KB3)}
 * E8   9 0  VALUE={PWR(1-EXP(-V(81)*(KC+KR1*V(82))/(KD+KR2*V(83))),KB4)*1.5708}
 * RE4  8 0  1
 * RE5 81 0  1
 * RE6 82 0  1
 * RE7 83 0  1
 * RE8  9 0  1 */
SPICE advancedPlateAndScreenCirV9Case2 () { 
}

WITH String optionalPlateSlope (String K) {
  if (KLAM  == 0 && KLAMG == 0) return "";
  if (KLAMG == 0) return "+KLAM*V(P,"+K+")";
  if (KLAM  == 0) return "*(1+KLAMG*V(P,"+K+"))";
  return "*(1+KLAMG*V(P,"+K+"))+KLAM*V(P,"+K+")";
}

WITH String optionalScreenSlope (String K) {
  if (KLAMG == 0) return "";
  return "/(1+KLAMG*V(P,"+K+"))";
}

WITH_SPICE_SUBCKT
/** 
  * **** ?NAME? ** Composite DHP with Advanced Grid Current **************
  * * Created on ?DATE? using paint_kip.jar
  * * www.dmitrynizh.com/tubeparams_image.htm
  * * Plate Curves image file: ?plateCurvesImageFilePath?
  * * Data source link: <plate curves URL>
  * *----------------------------------------------------------------------------------
  * .SUBCKT DHP_?NAME? P G K1 K2 G2 ; 1 2 3 4 5
  * + PARAMS: CCG=?CCG?P  CGP=?CGP?P CCP=?CCP?P RGI=?RGI? RFIL=?RFIL?
  * + MU=?MU? KG1=?KG1? KP=?KP? KVB=?KVB? KVC=?KVC? VCT=?VCT? EX=?EX? KG2=?KG2? KNEE=?KNEE?  
  * * Vp_MAX=?Vp_MAX? Ip_MAX=?Ip_MAX? Vg_step=?Vg_step? Vg_start=?Vg_start? Vg_count=?Vg_count?
  * * Rp=?Rp? Vg_ac=?Vg_ac? P_max=?P_max? Vg_qui=?Vg_qui? Vp_qui=?Vp_qui? UL=?UL? EG2=?EG2?
  * * X_MIN=?X_MIN? Y_MIN=?Y_MIN? X_SIZE=?X_SIZE? Y_SIZE=?Y_SIZE? FSZ_X=?FSZ_X? FSZ_Y=?FSZ_Y? XYGrid=?XYGrid?
  * * showLoadLine=?showLoadLine? showIp=?showIp? isDHP=?isDHP? isPP=?isPP? isAsymPP=?isAsymPP? isUL=?isUL? showDissipLimit=?showDissipLimit? 
  * * showIg1=?showIg1? isInputSnapped=?isInputSnapped?
  * * XYProjections=?XYProjections? harmonicPlot=?harmonicPlot? dissipPlot=?dissipPlot? 
  * * gridLevel2=?gridLevel2? 
  * *----------------------------------------------------------------------------------
  * RFIL_LEFT    K1  31 {RFIL/4}
  * RFIL_RIGHT   K2  41 {RFIL/4}
  * RFIL_MIDDLE1 31 34 {RFIL/4}
  * RFIL_MIDDLE2 34 41 {RFIL/4}
  * E11 32 0 VALUE={V(G2,31)/KP*LOG(1+EXP((P/MU+(VCT+V(G,31))/SQRT(KVB+V(G2,31)*V(G2,31)))*KP))} 
  * E12 42 0 VALUE={V(G2,41)/KP*LOG(1+EXP((P/MU+(VCT+V(G,41))/SQRT(KVB+V(G2,41)*V(G2,41)))*KP))} 
  * RE11 32 0 1G 
  * RE12 42 0 1G 
  * G11 P 31 VALUE={(PWR(V(32),EX)+PWRS(V(32),EX))/(2*KG1)*ATAN(V(P,31)/KNEE)}
  * G12 P 41 VALUE={(PWR(V(42),EX)+PWRS(V(42),EX))/(2*KG1)*ATAN(V(P,41)/KNEE)}
  * * Alexander Gurskii screen current, see audioXpress 2/2011
  * * todo: add dummy nodes as in IHP
  * G21 G2 31 VALUE={(PWR(V(32),EX)+PWRS(V(32),EX))/(2*KG2)*(KVC-ATAN(V(P,31)/KNEE))}
  * G22 G2 41 VALUE={(PWR(V(42),EX)+PWRS(V(42),EX))/(2*KG2)*(KVC-ATAN(V(P,41)/KNEE))}
  * RCP1 P 34 1G 
  * C1 G 34 {CCG} ; CATHODE-GRID 
  * C2 G P {CGP} ; GRID=PLATE 
  * C3 P 34 {CCP} ; CATHODE-PLATE 
  * RE2 G 0 1G
  * EGC1 81 0 VALUE={V(G,31)-VGOFF} ; POSITIVE GRID THRESHOLD
  * GG1 G 31 VALUE={0.5*(IGA+IGB/(IGC+V(P,31)))*(MU/KG1)*(PWR(V(81),IGEX)+PWRS(V(81),IGEX))}
  * EGC2 82 0 VALUE={V(G,41)-VGOFF} ; POSITIVE GRID THRESHOLD
  * GG2 G 41 VALUE={0.5*(IGA+IGB/(IGC+V(P,41)))*(MU/KG1)*(PWR(V(82),IGEX)+PWRS(V(82),IGEX))}
  * .ENDS 
  * *$
  * 
  */
SPICE DHP_KorenImproved_AdvancedGrid_Model () { 
  doLIN(MU,KG1,KP,KVB,VCT,EX,KNEE);
}

WITH String generateModelText () {
  if (isDHP == false) {
    return genKorenImprovedSpiceModel() + genKorenImprovedTriodeSpiceModel();
  }  else {
    if (gridLevel2)
      return DHP_KorenImproved_AdvancedGrid_Model();
    return genDHPKorenImprovedSpiceModel();
  }
}

WITH final double plateCurrentTriode (double Ep, double Eg, double kg) {
  double adj_e = KP * (1/MU + (VCT + Eg)/Math.sqrt(KVB + Ep*Ep));
  double E1 =
    (adj_e > 700)
    ? Ep/KP*adj_e // avoid overflow preserving precision
    : Ep/KP*LOG(1+EXP(adj_e));
  double Iplate = PWR_PWRS(E1,EX)/kg;
  return Iplate;
} 

// see https://en.wikipedia.org/wiki/Sigmoid_function
// https://en.wikipedia.org/wiki/Error_function
// https://www.desmos.com/calculator/rxyjk5jjvo
WITH double knee_func_erf (double x) {
  double xb1 = PWR(x, KB1);
  double xb2 = PWR(x, KB2);
  double xb3 = PWR(x, KB3);
  double power = -xb1*(KC+KR1*xb2)/(KD+KR2*xb3);
  return 1.5708*PWR((1-EXP(power)),KB4);
}

WITH double Eg_last = 1e9;

// these adjustments used to include current-proportional slope KLAMG*Eg and
// the kink, see https://www.desmos.com/calculator/64of4ge5ym where
// now the kink only
// the KNK part is current-proportional and KNG part is neg bias
// proportional.
// see also SPICE optionalPlateSlope, optionalScreenSlope
WITH double top_adjustments (double Ep, double Eg) {
  if (addKink == false) 
    return 1;
  // to do: improve this by relacing LIMIT with the approach from 
  // https://www.desmos.com/calculator/64of4ge5ym
  double magnitude = LIMIT(KNK-Eg*KNG,0,0.3);
  double dent_eq = // see https://www.desmos.com/calculator/64of4ge5ym
    -ATAN((Ep-KNPL)/KNSL)  // left side of 'trough' aka kink aka dip
    +ATAN((Ep-KNPR)/KNSR); // right side of 'trough' aka kink aka dip
  dent_eq *= magnitude;
  dent_eq +=1;
  return dent_eq;
}

WITH final double plateCurrent (double Ep, double Eg) {
  if (addLocalNFB) Eg += Ep*NFB; // apply local NFB
  if (currentTabIdx == 2 && doLIN && Eg_last != Eg) 
    plateCurrent_li_adjust(Ep, Eg);

  double e_screen = isUL? (Vp_qui + UL * (Ep - Vp_qui)) : EG2;
  double triode_ip = plateCurrentTriode(e_screen, Eg, KG1);
  return (advSigmoid
          ? (triode_ip * knee_func_erf(Ep/(
                                            //KNEE*(1- KVBG*(1-EXP(KVBGI*Eg)))
                                            // (KVBGI+triode_ip*KVBG/KNEE)/KNEE
                                            KNEE*(KVBGI+triode_ip*KG1*KVBG)///(KVBG+KVBGI)
                                            )
                                        ))
          : triode_ip * knee_simple(Ep))
    * top_adjustments(Ep, Eg) * (1 + KLAMG*Ep)  + KLAM*Ep;
}

WITH
/**
 *
 */
double gridCurrent_advanced (double Ep, double Eg) {
  Eg = Eg-VGOFF;
  double Igrid = (IGA+IGB/(IGC+Ep))*(MU/KG1)*PWR_PWRS(Eg,IGEX);
  return Igrid;
}


WITH
/**
 *
 */
double gridCurrent_Dx_RGI (double Ep, double Eg) {
  // these params define a very good aproximation of 
  // DX+RGI sucircuit as in:
  // D3 5 3 DX ; POSITIVE GRID CURRENT 
  // R1 2 5 {RGI} ; POSITIVE GRID CURRENT 
  // .MODEL DX D(IS=1N RS=1 CJO=10PF TT=1N) 
  // the params were found with PaintKIT.
  double ig_kp = 1280, ig_mu = 163.53, ig_vct = -0.00175, 
    ig_ex = 1.085, ig_kg1 = 0.00877;
  double vg1 =
    (Eg / ig_kp) *  
    LOG(1 + EXP(ig_kp * (1/ig_mu + ig_vct/Math.sqrt(Eg*Eg))));
  double ig = PWR_PWRS(vg1, ig_ex) / (RGI*ig_kg1);
  return ig;
}


WITH static int[] TR0 = {modelParamsYSpan()+110, // end of model sliders
                         40 // 3 rows of checkboxes
}, TR = 
  {70, 90, 
   TR0[0], TR0[0]+20, TR0[0]+40, 
   TR0[0]+TR0[1]+slidersYSpan(), TR0[0]+TR0[1]+slidersYSpan()+20,
   TR0[0]-(6*21)
};


WITH_TRACE 
double plateCurrent_a1 (double Ep, double Eg) {
  if (Eg > 0) return -1e9; // this value break the trace loop with given Eg
  return plateCurrent(Ep, Eg);
} 
LeftToRight Top Eg 
{toggle(showIp, true, 0, TR[0])
    //,color("Plate Current", 16, TR[0])
    ,color("Plate Current", 16) // abbrev trailing form
    ,toggle(isDHP, false, "DHP", 64, 8, "Generated Composite Directly Heated Model<br>with two pins for filament having resistance RFIL")
    ,toggle(isUL, {false, EG2_Label.setText("Vg2=gui"), EG2_Label.setText("Vg2")}, "UltraLinear", 156, TR[0], "Switch to Ultra-Linear connection mode<br>with ultra linear tap UL")
    ,toggle(addKink, false, "Kink", 120, TR[7], "Set  curve's 'kink' or 'dent'.<br>" + 
            "Use these additional controls to fit curves with visible 'kink' or 'dent'<br>" +
            "such as what is common for most tetrodes and due to secondary emission.")
    ,toggle(isTanhKnee, {false, advSigmoid(false)}, "Knee/Slope", 170, TR[7], "set sharper knee of the curves with fast transition to 'flat top',<br>" +
            "which is more suitable for many beam tetrodes.<br>" +
            "also, can be used to fit certain pentodes with very inclined front slope of the curves.<br>" +
            "Note: the slope is grid-bias-dependent (more inclined with more negative Vg).<br>" +
            "If this is not a good fit, you may try activating 'Adv Knee,Slope'")
    ,toggle(advSigmoid, {false, isTanhKnee(false)}, "Adv Knee,Slope", 270, TR[7], "pop up panel with additional controls")
    ,mirror(isPP && showLoadLine, abscissas_pix[3])
}

WITH_TRACE double plateCurrent_a2 (double Ep, double Eg) {
  if (Eg <= 0) return -1e9; // this value break the trace loop with given Eg
  return plateCurrent(Ep, Eg);
} 
LeftToRight Top Eg 
{//toggle(a2_showIp, true, 162, TR[0])
 displayIf(showIp)
   ,color("A2", 110, TR[0])
   ,mirror(isPP && showLoadLine, abscissas_pix[3])
}

// note: needs qui Ep and Eg
WITH double plateCurrentPP (double Ep, double Eg, double epqui, double egqui) {
    double i1 = plateCurrent(Ep, Eg);
    double i2 = 0;

    // changed to warning...
    if (2 * epqui < Ep) 
      System.out.println("-- (2 * epqui < Ep)!! " + Ep);
  
    // corrected formula
    i2 = plateCurrent(2 * epqui - Ep, 
		      egqui - (K_asym * // or 1
			       (Eg - egqui)));
  
    return i1 - i2;
}

WITH_TRACE double plateCurrentPP_trace (double Ep, double Eg) {
  if (showLoadLine) { // because we need Vp_qui and Vg_qui
    double i_plate = plateCurrentPP(Ep, Eg, Vp_qui, Vg_qui);
    return i_plate;
  }
  else
    return -1E9; // nothing
} 
LeftToRight Top Eg 
{toggle(isPP, false, 230, TR[4])
    ,color("PP", 18) // abbrev trailing form
    ,toggle(isAsymPP, false, "Asym", 230+50, TR[4], 70, "Asymmetric drive mode")
}

//was:   ,toggle(pushPullMode, false, "PP", 345, TR[0])

// future: combine plateCurrent_pos, plateCurrent_norm
// {toggle(showIp, true, "Plate Current")
//    ,"Colors:"
//    ,color("+ Vg")
//    ,color("0 or -")
// }

WITH_TRACE double gridCurrent (double Ep, double Eg) {
  if (gridLevel2)
    return gridCurrent_advanced(Ep, Eg);
  else // simple Dx+RGI model
    return gridCurrent_Dx_RGI(Ep, Eg);
} RightToLeft Top Eg 
{ toggle(showIg1, false, 0, TR[1])
    ,color("Grid Current", 16) // abbrev trailing form
    ,toggle(gridLevel2, false, "Level 2", 110, TR[1], "Advanced Grid Current.<br>"+
            "This is an extension of the classic Koren Improved Pentode's DX+RGI grid current.<br>" +
            "The Advanced grid current is based on Child's law and is Plate-voltage-dependent.<br>" +
            "Use the parameters on the pop-up box." )
    // ,toggle(subtractIg, false, "Ip-Ig")
    }

WITH_TRACE double dissipationLimit (double Ep, double Eg) {
  if (Eg != Vg_start) return -1e9; // this value break the trace loop with given Eg
  //return 0.005 * P_max;
  if (Ep < 1) return Ip_MAX;
  return P_max / Ep;
} LeftToRight  Bottom P_max 
{ toggle(showDissipLimit, true, 0, TR[2])
    ,color("Dissipation", 16) // abbrev trailing form
    ,mirror(isPP && showLoadLine, abscissas_pix[3])
    }


WITH_TRACE double loadLine (double Ep, double Eg) {
  if (isPP || Eg != Vg_start) return -1e9; // this value break the trace loop with given Eg
  return Ip_qui + (Vp_qui - Ep) / Rp;
} LeftToRight Bottom Rp 
{ toggle(showLoadLine, blankBGMode, // on when in blank mode
	 0, TR[3])
    ,color("Loadline for load Rp", 16) // abbrev trailing form
    ,toggle(isInputSnapped, true, "Vg span", 156, TR[3])
    ,toggle(addLocalNFB, false, "NFB", 156, TR[2], "Add Local NFB")
    ,toggle(loadAndFit, false, "unlock SPICE params", 230, TR[2], 
	    "Allow changes to SPICE model prameters MU KG1 etc <br>when fitting the operation point and loadline.")
    ,toggle(dissipPlot, false, "Dissipation Plot", 230, TR[3], 
	    "Display small plot that shows plate dissipation<br>for Vg ranging from 0 to Vin.")
    } OVERWRITING addLocalNFB

PROLOGUE checkboxStateChanged /*(Checkbox checkbox, ItemEvent e)*/ {
  // out of vogue... if (isPP && checkbox == showLoadLine_ToggleBox) isPP(false);

  if (showLoadLine && checkbox == isPP_ToggleBox) { // adjust Y_SIZE
    if (isPP) {
      adjustXY(X_SIZE, 2*Y_SIZE);
      xMarksOnWhiteBG = false ;
    } else {
      adjustXY(X_SIZE, Y_SIZE/2);
      parseVarValPair("Vp_MAX", ""+2*Vp_qui);
      xMarksOnWhiteBG = true;
    }
  }    
}


WITH_DRAWINGS void XYProjections (Graphics g) {
  if (showLoadLine) {
    drawLoadlinePointProjections(g, abscissas[0], ordinates[0]);
    drawLoadlinePointProjections(g, abscissas[3], ordinates[3]);
    drawLoadlinePointProjections(g, abscissas[6], ordinates[6]);
  }
} false "XYProj" 90 TR[4]
"X,Y axis projections of the quiescent point<br> and loadline endpoints."

// placed ahead of the below drawings loadline_calcs so that it is called lexically after it
WITH_DRAWINGS void harmonicPlot (Graphics g) {
  if (showLoadLine) drawHarmonics(g);
} true "HPlot"  156 TR[4]
"Plot displaying harmonics in %.<br>" + 
 "Alone the X axis is input signal from zero to Vin amplitude<br>" + 
 "aisY is distortion in %."

WITH_DRAWINGS void loadline_calcs (Graphics g) {
  if (showLoadLine) { 
    double start = 0, end = 0;
    if (isInputSnapped) {
      start = Vg_start; end = Vg_start - Vg_step * (Vg_count-1);
    }
    else {
      start = Vg_qui+Vg_ac;
      end = Vg_qui-Vg_ac;
    }
    sevenOrdinates(start, end, harmonicPlot || dissipPlot);
    if (dissipPlot) drawPowerDissipationPlot(harmonicPlot, g);
  }
} true "   Results" 0-20  TR[4] // offscreen! it really should always be on when loadlines are on
"Compute and display quiescent point parameters,<br>loadling parameters, harmonics distortion data and dissipation."


// assume Ep > 0
WITH_TRACE double screenCurrent (double Ep, double Eg) {
  double e_screen = isUL? (Vp_qui + UL * (Ep - Vp_qui)) : EG2;
  double triode_ip = plateCurrentTriode(e_screen, Eg, 1);
  double knee_coeff = advSigmoid 
    ? knee_func_erf(Ep/(KNEE*(KVBGI+triode_ip*KVBG)///(KVBG+KVBGI)
                        )
                    )
    : knee_simple(Ep);
  
  double Ip_dip = knee_coeff*(triode_ip/KG1)*(1 - top_adjustments(Ep, Eg));
  triode_ip = triode_ip/KG2;
  return triode_ip * (KVC - knee_coeff)/(1 + KLAMG*Ep)  + Ip_dip;

} LeftToRight Top Eg
{ toggle(showIg2, false, 230, TR[1], "Screen Grid (aka \"grid #2\") Current, mA.")
    ,color("Screen Current", 16) // abbrev trailing form
    }

WITH_TRACE double combinedCurrent (double Ep, double Eg) {
  // debugger for triode-strapped kg1 formula
  //  double triode_ip = plateCurrentTriode(Ep, Eg, (4*KG1*KG2/((2*KVC-3.1416)*(2*KG1+3.1416*KG2))));

  double triode_ip = screenCurrent(Ep, Eg) + plateCurrent(Ep, Eg);
  return triode_ip;
} LeftToRight Top Eg
{ toggle(showIpAndIg2, false, 0, TR[6], "Combined Plate plus Screen Grid (aka \"grid #2\") Current, mA.")
    ,color("Ip+Ig2", 16) // abbrev trailing form
    }

WITH_TRACE double triodeStrappedCurrent (double Ep, double Eg) {
  // this is true for strict KIP only
  // return plateCurrentTriode(Ep, Eg, KG2/(KVC-Math.PI/2)) + plateCurrentTriode(Ep, Eg, 2*KG1/Math.PI);
  double saved_UL = UL; boolean saved_isUL = isUL;
  UL = 1; isUL = true;
  double current = combinedCurrent(Ep, Eg);
  UL = saved_UL; isUL = saved_isUL;
  return current;
} LeftToRight Top Eg
{ toggle(Ip_triodeStrapped, false, 70, TR[6], "Current by this tube Triode-strapped (plate and screen shorted), mA.<br>"+
    "Computed as the sum Ip + Is when UL=1."+
    "See also [Triode Model Ip] trace.")
    ,color("Triode-Strapped Current", 16) // abbrev trailing form
    }

WITH_TRACE double triodeModelPlateCurrent (double Ep, double Eg) {
  // this is true for strict KIP only
  // return plateCurrentTriode(Ep, Eg, KG2/(KVC-Math.PI/2)) + plateCurrentTriode(Ep, Eg, 2*KG1/Math.PI);
  return plateCurrentTriode(Ep, Eg, 4*KG1*KG2/((2*KVC-Math.PI)*(2*KG1+Math.PI*KG2)));
} LeftToRight Top Eg
{ toggle(Ip_triodeModel, false, 250, TR[6], "Plate Current, mA, for the Derived Triode Model.<br>"+
    "Such triode's Kg1 is roughly 0.7*Kg1||Kg2,"+
    "or precisely 4Kg1Kg2 / ((2Kvc-Pi)(2Kg1+PiKg2))<br>" +
    "This model is complimentary and is always generated in addition to the<br>" +
    "pent/tetr model when the [Model] button is pressed.<br>" +
    "Note that for simplicity and efficiency, this model omits the top inclination/slope (KLAM, KLAMG),<br>" +
    "compare to [Triode-strapped Current] trace.")
    ,color("Triode Model Ip", 16) // abbrev trailing form
    }


// WITH_DRAWINGS void showIVGrid (Graphics g) {
// } showPlotGridlines "I/V Gridlines" 200 8


// *** PARAMS **** 

// WITH_SLIDER note: the 1arg is java variable; many are  used in code above

// Triode params
WITH_MODEL_PARAM_GROUP TRIODE_PARMS "Triode params" "SPICE Model parameters that are common for triode<br>" +
"and tetrode/pentode operation<br>" +
"Tip: set these first with a triode-strapped BG image,<br>using UltraLinear and UL=1, then come back to pentode param fitting";
WITH_MODEL_PARAM MU 10 true ;
WITH_MODEL_PARAM KG1 2200  true ;
WITH_MODEL_PARAM KP  18.77 true ;
WITH_MODEL_PARAM KVB 12.0  true ;
WITH_MODEL_PARAM VCT 0.2 true ;
WITH_MODEL_PARAM EX 1.4 true ;

// Grid Models
// Level 1 (Koren Dx augmented with VGOFF)
WITH_MODEL_PARAM RGI {2000, "For Grid Current when grid is positive."} showIg1 && !gridLevel2 ;

// pent
WITH_MODEL_PARAM_GROUP PENT_PARAMS "Pent/Tetr Params" "SPICE Model parameters that are specific to<br>" +
"tetrode/pentode behavior";
WITH_MODEL_PARAM KG2 4200 true ;
WITH_MODEL_PARAM KLAM 2e-6 & "Sets a slight linear slope of the top of the curves<br>" +
"This is an extension of the classic Koren Improved Pentode.<br>" +
"The slope is independent of grid voltage and is meant to be slight.<br>" +
"To disable, and remove from the SPICE MODEL, set this param box to O (zero)"
 true 
if (KLAM < 0) KLAM_TF.setText(format_double(KLAM=KLAM_middle=0));

WITH_MODEL_PARAM KLAMG {3e-4,"Simple Grid-dependent linear slope of the top of the curves<br>" +
    "This is an extension of the classic Koren Improved Pentode.<br>" +
    "To disable, and remove from the SPICE MODEL, set this param box to O (zero)"} true
    if (KLAMG < 0) KLAMG_TF.setText(format_double(KLAMG=KLAMG_middle=0));

// pentode ATAN knee params
WITH_MODEL_PARAM KNEE 12 | "Pentode Knee Parameter. It is the parameter KVB in the original Koren Improved Model" true ;
// WITH_MODEL_PARAM KLAMEX 1.3 true ;

// The 'action' for KVC limits KVC to values above Pi/2
WITH_MODEL_PARAM KVC 2.57 | "This model parameter controls to what degree the knee bend of the plate lines is mirrored by the<b>" +
"screen lines. When KVC=1.57, its lowest possible value, the grid lines<br>" +
"mirror the plate lines, scaled according to KG1 and KG2 value; when KG1-KG2, then <br>" + 
"each screen line starts at some Iso current at Vp=0 and approaches zero when Vp is very high;<br>" +
"when it is PI, each line starts at some Iso current at Vp=0 and ends at Iso/2 when Vp is very high." 
(showIg2||isUL||showIpAndIg2) if (KVC < 1.5708) SET_PARAM KVC 1.5708; // KVC_TF.setText(format_double(KVC = 1.5708));

// Tahn knee is sharper and better for beam tetrodes
// atan still defines the top but is shifted
WITH_PARAMS_WINDOW CompositeKnee "Knee with beam tetrode corner" isTanhKnee {KNEE2, KNEX};
WITH_MODEL_PARAM KNEE2 20 | "this controls the knee. The main window parameter KNEE is still in effect,<br>" +
"and controls curving on the top of the knee." true;
WITH_MODEL_PARAM KNEX 30 | "by how many volts the knee controlled by parameter KNEE is shifted to the left." true;

// pentode signoid knee params
// https://en.wikipedia.org/wiki/Error_function
// WITH_MODEL_PARAM KA 0.07 true ;
//                                            alt init vals   V1
WITH_PARAMS_WINDOW AdvKnee "Advanced Knee Params" advSigmoid 
{FRONT_SL_KNEE, KD , KC , KR1, KR2, 
    ADV_S, KVBG, KB1, KB2, KB3, KB4, KVBGI};

WITH_MODEL_PARAM_GROUP FRONT_SL_KNEE "Front slope and knee radius" "Tip: use these in conjunction with KNEE";
WITH_MODEL_PARAM KD  0.44   true ;             //              2.4 
WITH_MODEL_PARAM KC  0.14   true ;             //              0.3  
WITH_MODEL_PARAM KR1 0.0013 true ;             //              0.01 
WITH_MODEL_PARAM KR2 0.064  true ;             //              0.04 

WITH_MODEL_PARAM_GROUP ADV_S "More advanced S-curve shape params" "These parameters help to dial into more elaborate shape.<br>" +
"Tip: modify these only if using KD,KC, KR1 and LR2 is not enough";
// pentode slope params
WITH_MODEL_PARAM KVBG  0.015 true ; // 29/KG1
WITH_MODEL_PARAM KVBGI 0.001 true ;
WITH_MODEL_PARAM KB1 2 true ;                  //              2 
WITH_MODEL_PARAM KB2 2 true ;                  //              2 
WITH_MODEL_PARAM KB3 2 true ;                  //              2 
WITH_MODEL_PARAM KB4 0.5  true ;               //              0.5  

WITH_PARAMS_WINDOW DENT "Control curve's 'kink'" addKink 
//"Use these to fit curves with visible 'kink' or 'dent'.<br>" +
//"Tip1 (Impact on complexity of generated SPICE Model): as long as KNK=KNG=0, there is no overhead - <br>" + 
//"kink complexity is <b>not</b> included into the SPICE model" +
//"Tip2: As experimental feature, Kink remains active even for the default Koren-style Knee (Atan)." 
{KNK, KNG, KNPL, KNPR, KNSL, KNSR};

WITH_MODEL_PARAM KNK -0.044 | "Tetrode 'kink' amount portion proportional to plate current<br>" +
"Note: what gets subtracted from Ip, gets added to Ig2..." true ; 
WITH_MODEL_PARAM KNG 0.006  | "Tetrode 'kink'amount portion proportional to negative Vg<br>" +
"Hint: to define cerain negative bias threshold Vt for this, set KNK  -10 * KNG<br>" +
"Example: to have the kink to grow on curves with bias > -10V:  KNK = -10 * KNG"  true ; 
WITH_MODEL_PARAM KNPL 50  | "Vp where the dip starts"  true if (KNPL > KNPR) KNPL_TF.setText(format_double(KNPL = KNPR));
WITH_MODEL_PARAM KNSL 11  | "Vp where the dip ends" true ; 
WITH_MODEL_PARAM KNPR 120 | "left-side slope of the dip (usually shallow)"  true if (KNPR < KNPL) KNPR_TF.setText(format_double(KNPR = KNPL));
WITH_MODEL_PARAM KNSR 29  | "right-side slope of the dip (often shallow but sometimes can be steep)" true ; 


// Power
WITH_SLIDER P_max "Pmax W" 25 showDissipLimit ;

// Load
WITH_SLIDER Rp     "Rp ohm" 1400 showLoadLine ;
WITH_SLIDER Vg_ac  "Vin ampl" 20 showLoadLine && !isInputSnapped ;
// WITH_SLIDER Vg_qui Vg_start - (Vg_step * Vg_count)/2 showLoadLine ;
WITH_SLIDER Vg_qui "Vg qui" 0-20 showLoadLine && !isInputSnapped ;
WITH_SLIDER Vp_qui "Vp qui" 300 showLoadLine 
 if (isPP) parseVarValPair("Vp_MAX", ""+2*Vp_qui);

// Level 2 
WITH_PARAMS_WINDOW AdvGrid "Advanced Grid Params" gridLevel2 {VGOFF,IGA,IGB,IGC,IGEX};

WITH_MODEL_PARAM VGOFF 0-0.6 showIg1 && gridLevel2 ;
WITH_MODEL_PARAM IGA 0.001  showIg1 && gridLevel2 ;
WITH_MODEL_PARAM IGB 0.3    showIg1 && gridLevel2 ;
WITH_MODEL_PARAM IGC 8.0    showIg1 && gridLevel2 ;
WITH_MODEL_PARAM IGEX 2.0   showIg1 && gridLevel2 ;

//WITH double VGOFF  = 0-0.6 ; // showIg1 && gridLevel2 ;
//WITH double IGA    = 0.001 ; //  showIg1 && gridLevel2 ;
//WITH double IGB    = 0.3   ; //  showIg1 && gridLevel2 ;
//WITH double IGC    = 8.0   ; //  showIg1 && gridLevel2 ;
//WITH double IGEX   = 2.0   ; // showIg1 && gridLevel2 ;

WITH_SLIDER K_asym  "Asym" 1.0   isPP && isAsymPP ;

WITH_SLIDER UL "UL" 0.43 isUL;
WITH_SLIDER EG2 "Vg2" 300 !isUL; // see toggle isUL actions

WITH_SLIDERS_WINDOW LOCAL_NFB "Apply Local NFB" addLocalNFB {NFB};
WITH_SLIDER NFB "NFB" 0.04 addLocalNFB; OVERWRITING NFB

WITH_ENHANCED_MOUSE_SUPPORT showLoadLine isPP

EPILOGUE mouseWheelMovedR1C1 if (modelParamPrecondition("")) { MU(MU * (100 + 3 * rot) / 100.0D); }
EPILOGUE mouseWheelMovedR1C2 if (modelParamPrecondition("")) { MU(MU * (100 + 3 * rot) / 100.0D); KG1(KG1 * (100 - 3 * rot) / 100.0D);}
EPILOGUE mouseWheelMovedR1C3 if (modelParamPrecondition("")) { KG1(KG1 * (100 + 3 * rot) / 100.0D); }
EPILOGUE mouseWheelMovedR2C1 if (modelParamPrecondition("")) { 
  if (isUL) KVB(KVB * (100 + 3 * rot) / 100.0D); 
  else KNEE(KNEE * (10 + 3 * rot) / 10.0D); 
 }
EPILOGUE mouseWheelMovedR2C2 if (modelParamPrecondition("")) { EX(EX * (100 + 3 * rot) / 100.0D); }
EPILOGUE mouseWheelMovedR2C3 if (modelParamPrecondition("")) { KP(KP * (100 + 3 * rot) / 100.0D); }



// if doLIN is true, restore some model params as well
WITH_TAB_SWITCH doLIN(MU, KG1, KP, KVB, EX);
EPILOGUE switchToImageAndConfigAux {
  isUL(isUL = v.isUL);
  UL(v.UL);
  showIg2(v.showIg2);
  EG2(v.EG2);
  showIpAndIg2(v.showIpAndIg2);
  Ip_triodeStrapped(v.Ip_triodeStrapped);
  Ip_triodeModel(v.Ip_triodeModel);
}


EPILOGUE fillCentralPanel {
  gridLevel2_ToggleBox.setEnabled(showIg1);
  isAsymPP_ToggleBox.setEnabled(isPP);

  isInputSnapped_ToggleBox.setEnabled(showLoadLine);
  addLocalNFB_ToggleBox.setEnabled(showLoadLine);
  loadAndFit_ToggleBox.setEnabled(showLoadLine);

  dissipPlot_ToggleBox.setEnabled(showLoadLine);

  harmonicPlot_ToggleBox.setEnabled(showLoadLine);
  harmonicPlot_CounterButton.setEnabled(showLoadLine);
  
  loadline_calcs_ToggleBox.setEnabled(showLoadLine);
  loadline_calcs_CounterButton.setEnabled(showLoadLine);
  loadline_calcs = showLoadLine; // sync the state

  XYProjections_ToggleBox.setEnabled(showLoadLine);
  XYProjections_CounterButton.setEnabled(showLoadLine);
  isPP_ToggleBox.setEnabled(showLoadLine);

  advSigmoid_ToggleBox.setEnabled(!showLoadLine || loadAndFit);
  isTanhKnee_ToggleBox.setEnabled(!showLoadLine || loadAndFit);
  addKink_ToggleBox.setEnabled(!showLoadLine || loadAndFit);

  // showIpAndIg2_ToggleBox.setEnabled(isUL);

  // MU_SB.setVisible(false);
  // MU_TF.setVisible(false);
  // MU_Label.setVisible(false);

  // KG1_SB.setEnabled(!lockSPICE);
  // KG1_TF.setEnabled(!lockSPICE);
  // KP_SB.setEnabled(!lockSPICE);
  // KP_TF.setEnabled(!lockSPICE);
  // KVB_SB.setEnabled(!lockSPICE);
  // KVB_TF.setEnabled(!lockSPICE);
  // VCT_SB.setEnabled(!lockSPICE);
  // VCT_TF.setEnabled(!lockSPICE);
  // EX_SB.setEnabled(!lockSPICE);
  // EX_TF.setEnabled(!lockSPICE);

}

PROLOGUE modelParamPrecondition {
  if (showLoadLine && loadAndFit == false) return false;
}

// Java Class definition
DEF_MODEL_PARAM_FINDER PaintKIP;

